package flink.zanxiangnet.ad.monitoring.process;

import com.aliyun.odps.Instance;
import com.aliyun.odps.Odps;
import com.aliyun.odps.account.Account;
import com.aliyun.odps.account.AliyunAccount;
import com.aliyun.odps.data.Record;
import com.aliyun.odps.task.SQLTask;
import flink.zanxiangnet.ad.monitoring.maxcompute.MaxComputeLog;
import flink.zanxiangnet.ad.monitoring.pojo.entity.AdDataOfMinuteODS;
import flink.zanxiangnet.ad.monitoring.pojo.entity.AdStatOfDayDWD;
import flink.zanxiangnet.ad.monitoring.pojo.entity.AdStatOfMinuteDWD;
import flink.zanxiangnet.ad.monitoring.pojo.properties.ApplicationProperties;
import flink.zanxiangnet.ad.monitoring.util.DateUtil;
import flink.zanxiangnet.ad.monitoring.util.JsonUtil;
import org.apache.flink.api.common.state.MapState;
import org.apache.flink.api.common.state.MapStateDescriptor;
import org.apache.flink.api.common.state.ValueStateDescriptor;
import org.apache.flink.configuration.Configuration;
import org.apache.flink.streaming.api.functions.windowing.ProcessWindowFunction;
import org.apache.flink.streaming.api.windowing.windows.TimeWindow;
import org.apache.flink.util.Collector;

import java.time.LocalDate;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;

public class AdMinuteDWDProcess extends ProcessWindowFunction<AdDataOfMinuteODS, AdStatOfMinuteDWD, Long, TimeWindow> {
    // 对象锁，防止MaxCompute的 Tunnel对象多次初始化
    private static final Object DUMMY_LOCK = new Object();
    private Odps odps;
    private long lastClearTime = System.currentTimeMillis();
    private final Map<Long, String> lastAdDayQueryDay = new ConcurrentHashMap<>();
    // 历史的天数据
    private final Map<Long, AdStatOfDayDWD> historyDayData = new ConcurrentHashMap<>();
    // 之前聚合的昨天的数据
    private final Map<Long, Map<String, AdStatOfMinuteDWD>> historyMinuteData = new ConcurrentHashMap<>();
    // 前 5分钟聚合的数据
    private final Map<Long, Map<String, AdStatOfMinuteDWD>> lastReduceData = new ConcurrentHashMap<>();

    @Override
    public void open(Configuration conf) {
        Map<String, String> params = getRuntimeContext()
                .getExecutionConfig()
                .getGlobalJobParameters()
                .toMap();
        Account account = new AliyunAccount(params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_ID),
                params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_KEY));
        odps = new Odps(account);
        odps.getRestClient().setRetryLogger(new MaxComputeLog());
        odps.setEndpoint(params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_ENDPOINT));
        odps.setDefaultProject(params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_PROJECT_NAME));
    }

    @Override
    public void process(Long elementCount, ProcessWindowFunction<AdDataOfMinuteODS, AdStatOfMinuteDWD, Long, TimeWindow>.Context context,
                        Iterable<AdDataOfMinuteODS> iterable, Collector<AdStatOfMinuteDWD> collector) throws Exception {
        MapState<String, AdStatOfMinuteDWD> historyMinuteState = context.windowState().getMapState(new MapStateDescriptor<>("historyMinuteData", String.class, AdStatOfMinuteDWD.class));
        LocalDateTime beginTime = DateUtil.milliToLocalDateTime(context.window().getStart());
        String statDay = DateUtil.formatLocalDate(beginTime.toLocalDate());
        String reduceTimeKey = DateUtil.formatOfMinuteReduce(beginTime);
        int hour = beginTime.getHour();
        long now = System.currentTimeMillis();
        AdDataOfMinuteODS statODS = null;
        List<AdDataOfMinuteODS> adDataOfMinuteODSList = new ArrayList<>(24);
        System.out.println("windowCount：" + adDataOfMinuteODSList);
        for (AdDataOfMinuteODS adDataOfMinuteODS : iterable) {
            adDataOfMinuteODSList.add(adDataOfMinuteODS);
            if (adDataOfMinuteODS.getHour() != hour) {
                continue;
            }
            statODS = adDataOfMinuteODS;
            break;
        }
        if (statODS == null) {
            return;
        }

        // 每 12小时清理一次数据
        if (now - lastClearTime > 12 * 60 * 60 * 1000) {
            synchronized (DUMMY_LOCK) {
                if (now - lastClearTime > 12 * 60 * 60 * 1000) {
                    clearData();
                    lastClearTime = now;
                }
            }
        }
        Long adId = statODS.getAdId();

        // 昨天聚合的数据
        Map<String, AdStatOfMinuteDWD> historyMinuteMapping = historyMinuteData.computeIfAbsent(adId, key -> new HashMap<>());
        AdStatOfMinuteDWD yesterdayMinuteDWD = historyMinuteMapping.get(DateUtil.formatLocalDate(beginTime.minusDays(1L).toLocalDate()));
        // 之前的数据
        String lastQueryDay = lastAdDayQueryDay.get(adId);
        if (lastQueryDay == null || !lastQueryDay.equals(statDay)) {
            // 往前找 30天
            LocalDate endDay = beginTime.minusDays(2L).toLocalDate(), beginDay = endDay.minusDays(30L);
            Instance instance = SQLTask.run(odps, "SELECT * FROM ad_stat_of_day_dwd WHERE stat_day >= \"" + DateUtil.formatLocalDate(beginDay) + "\" AND stat_day <= \"" + endDay + "\" AND ad_id = " + adId + ";");
            instance.waitForSuccess();
            List<Record> records = SQLTask.getResult(instance);
            List<AdStatOfDayDWD> temp = records.stream().map(AdStatOfDayDWD::byMaxCompute).sorted((val1, val2) -> {
                if (val1.getStatDay().equals(val2.getStatDay())) {
                    return new Long(val1.getCreateTime().getTime() - val2.getCreateTime().getTime()).intValue();
                }
                return DateUtil.parseLocalDate(val1.getStatDay()).compareTo(DateUtil.parseLocalDate(val2.getStatDay()));
            }).collect(Collectors.toList());
            if(!temp.isEmpty()) {
                historyDayData.put(adId, temp.get(temp.size() - 1));
            }
            System.out.println(JsonUtil.toString("历史天数据：" + historyDayData));
            lastAdDayQueryDay.put(adId, statDay);
        }
        AdStatOfDayDWD beforeYesterdayDayDWD = historyDayData.get(adId);
        Map<String, AdStatOfMinuteDWD> lastReduceMapping = lastReduceData.computeIfAbsent(adId, key -> new HashMap<>());
        // 聚合当天的全部数据
        AdStatOfMinuteDWD newAdStat = AdStatOfMinuteDWD.reduce(beforeYesterdayDayDWD, yesterdayMinuteDWD, adDataOfMinuteODSList, statODS, lastReduceMapping.get(DateUtil.formatOfMinuteReduce(beginTime.minusMinutes(5L))), now);
        collector.collect(newAdStat);
        System.out.println("输出：" + JsonUtil.toString(newAdStat));
        lastReduceMapping.put(reduceTimeKey, newAdStat);
        historyMinuteMapping.put(statDay, newAdStat);
    }

    @Override
    public void clear(ProcessWindowFunction<AdDataOfMinuteODS, AdStatOfMinuteDWD, Long, TimeWindow>.Context context) throws Exception {
        System.out.println("窗口关闭");
    }

    private void clearData() {
        lastAdDayQueryDay.entrySet().removeIf(entry -> DateUtil.parseLocalDate(entry.getValue()).compareTo(LocalDate.now().minusDays(3L)) <= 0);
        historyDayData.entrySet().removeIf(entry -> DateUtil.parseLocalDate(entry.getValue().getStatDay()).compareTo(LocalDate.now().minusDays(50L)) <= 0);
        historyMinuteData.forEach((key, value) -> {
            if (value != null) {
                value.entrySet().removeIf(entry2 -> DateUtil.parseLocalDate(entry2.getKey()).compareTo(LocalDate.now().minusDays(3L)) <= 0);
            }
        });
        historyMinuteData.entrySet().removeIf(entry -> entry.getValue() == null || entry.getValue().isEmpty());

        lastReduceData.forEach((key, value) -> {
            if (value != null) {
                value.entrySet().removeIf(entry -> DateUtil.parseOfMinuteReduce(entry.getKey()).compareTo(LocalDateTime.now().minusHours(1)) <= 0);
            }
        });
        lastReduceData.entrySet().removeIf(entry -> entry.getValue() == null || entry.getValue().isEmpty());
    }
}
